/**
  *  \file server/session.cc
  *  \brief Router: Session Implementation
  */

#include <vector>
#include <cstring>
#include <cctype>
#include <sys/wait.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include "server/session.h"
#include "server/request.h"
#include "server/filenotify.h"

bool
isPOST(const char* command)
{
    return std::toupper(uint8_t(command[0])) == 'P'
        && std::toupper(uint8_t(command[1])) == 'O'
        && std::toupper(uint8_t(command[2])) == 'S'
        && std::toupper(uint8_t(command[3])) == 'T'
        && command[4] == ' ';
}

bool
isSAVE(const char* command)
{
    return std::toupper(uint8_t(command[0])) == 'S'
        && std::toupper(uint8_t(command[1])) == 'A'
        && std::toupper(uint8_t(command[2])) == 'V'
        && std::toupper(uint8_t(command[3])) == 'E'
        && command[4] == 0;
}

static void
parseArgs(string_t name, std::vector<char*>& out)
{
    string_t::size_type p = name.find_first_not_of(" \t");;
    while (p < name.size()) {
        string_t::size_type q = name.find_first_of(" \t", p);
        string_t::size_type n = (q == name.npos ? name.size() - p : q - p);
        out.push_back(0);
        char* e = new char[n+1];
        name.copy(e, n, p);
        e[n] = 0;
        out.back() = e;
        if (q == name.npos)
            break;
        p = name.find_first_not_of(" \t", q);
    }
}

/** Check for conflict marker. */
static bool
isConflictMarker(const char* p)
{
    return p != 0
        && p[0] == '-'
        && (p[1] == 'R' || p[1] == 'W');
}

static bool
isSame(const char* a, const char* b, bool bIsWild)
{
    std::size_t an = std::strlen(a);
    std::size_t bn = std::strlen(b);
    if (bIsWild && bn > 0 && b[bn-1] == '*') {
        return an >= bn-1
            && std::memcmp(a, b, bn-1) == 0
            && (an == bn-1
                || a[bn-1] == '/');
    } else {
        return an == bn
            && std::memcmp(a, b, an) == 0;
    }
}

/** Check for conflict. Conflict resolution uses "reader/writer lock" terminology.
    Each session can be associated with a set of keywords, starting with "-R" or "-W"
    (they need not have a real-world meaning, i.e. they need not imply that someone
    is reading or writing).

    Any number with an identical "-R" keyword can coexist, but if there is a "-W"
    session, it must be the only one. A "-W" session wins over a "-R" session.
    Two "-W" sessions can be resolved either way ("refuse new session because old
    one still active" vs. "opening a new session closes old one").

    \param a,b Keywords
    \param aWinsTie if true, a wins a W/W tie. If false, b wins a W/W tie.
    \retval <0 if b wins
    \retval =0 if no conflict
    \retval >0 if a wins */
static int
checkConflict(const char* a, const char* b, bool aWinsTie, bool bIsWild)
{
    if (isConflictMarker(a) && isConflictMarker(b) && (a[1] == 'W' || b[1] == 'W') && isSame(a+2, b+2, bIsWild)) {
        /* This is a conflict */
        if (a[1] == 'W' && (b[1] != 'W' || aWinsTie)) {
            return +1;
        } else {
            return -1;
        }
    } else {
        /* Not a conflict */
        return 0;
    }
}

/** Constructor.
    \param name Name of session (=argv for server)
    \param id Session number */
RouterSession::RouterSession(string_t name, uint32_t id)
    : pid(0),
      readfd(-1),
      writefd(-1),
      session_id(id),
      last_access(time(0)),
      name(name),
      used(false),
      modified(false)
{
    args.push_back(0);
    parseArgs(name, args);
    args.push_back(0);
}

RouterSession::~RouterSession()
{
    stop();
    for (std::vector<char*>::iterator i = args.begin(); i != args.end(); ++i)
        delete[] *i;
}

void
RouterSession::maybeNotifyFileServer()
{
    for (std::vector<char*>::const_iterator i = args.begin(); i != args.end(); ++i) {
        if (*i != 0 && strncmp(*i, "-WDIR=", 6) == 0) {
            notifyFileServer("FORGET", *i+6);
        }
    }
}

/** Check for conflict.
    \param other Other session
    \return true iff this session conflicts with the other one */
int
RouterSession::checkConflict(const RouterSession& other, bool thisWinsTie) const
{
    for (std::vector<char*>::const_iterator i = args.begin(); i != args.end(); ++i)
        if (int result = other.checkConflict(*i, thisWinsTie, false))
            return result;
    return 0;
}

/** Check for conflict.
    \param query Keyword ("-Rfoo", "-Wfoo")
    \return true iff this session conflicts with this keyword */
int
RouterSession::checkConflict(const char* query, bool thisWinsTie, bool queryIsWild) const
{
    if (!::isConflictMarker(query))
        return 0;
    for (std::vector<char*>::const_iterator i = args.begin(); i != args.end(); ++i)
        if (int result = ::checkConflict(*i, query, thisWinsTie, queryIsWild))
            return result;
    return 0;
}

/** Start session.
    \param server Name of server binary
    \return true on success, false on failure */
bool
RouterSession::start(const char* server)
{
    /* don't do anything if session is already started */
    if (pid != 0)
        return true;

    /* make pipes */
    enum { Read, Write };
    int tochild[2], fromchild[2];
    if (pipe(tochild) != 0) {
        perror("pipe");
        return false;
    }
    if (pipe(fromchild) != 0) {
        perror("pipe");
        close(tochild[0]);
        close(tochild[1]);
        return false;
    }

    /* make child */
    pid_t child = fork();
    if (child < 0) {
        perror("fork");
        close(tochild[0]);
        close(tochild[1]);
        close(fromchild[0]);
        close(fromchild[1]);
        return false;
    }
    if (child == 0) {
        /* I am the child */
        dup2(tochild[Read], 0);
        dup2(fromchild[Write], 1);
        dup2(fromchild[Write], 2);
        close(tochild[0]);
        close(tochild[1]);
        close(fromchild[0]);
        close(fromchild[1]);
        args[0] = const_cast<char*>(server);
        execv(server, &args[0]);
        _exit(1);
    } else {
        /* I am the parent */
        this->readfd = fromchild[Read];
        this->writefd = tochild[Write];
        this->pid = child;
        close(fromchild[Write]);
        close(tochild[Read]);
        fcntl(readfd, F_SETFD, FD_CLOEXEC);
        fcntl(writefd, F_SETFD, FD_CLOEXEC);

        /* Wait for child to start up. It will write a "hello" message
           with a "100" code, or some error messages. */
        answer.clear();
        char* startup;
        while ((startup = answer.readLine()) == 0) {
            if (answer.readMore(readfd) != Request::sSuccess)
                break;
        }
        if (startup != 0 && std::strncmp(startup, "100", 3) == 0) {
            /* Looks like a success message */
            printf("session %d [%d]: child started for '%s':\n", int(session_id), int(pid), name.c_str());
            return true;
        } else {
            /* Looks like a failure message */
            printf("session %d [%d]: child failed to start for '%s', startup message follows:\n", int(session_id), int(pid), name.c_str());
            while (startup != 0) {
                printf("session %d [%d]:   %s\n", int(session_id), int(pid), startup);
                while ((startup = answer.readLine()) == 0
                       && answer.shift()
                       && answer.readMore(readfd) == Request::sSuccess)
                { }
            }
            stop();
            return false;
        }
    }
}

/** Stop this session. */
void
RouterSession::stop()
{
    /* no need to do anything if session is already stopped */
    if (pid == 0)
        return;

    /* terminate it by closing its stdin */
    printf("session %d [%d]: stopping...\n", int(session_id), int(pid));
    close(writefd);

    /* satisfy possibly pending output */
    do {
        answer.clear();
    } while (answer.readMore(readfd) == Request::sSuccess);
    close(readfd);

    /* terminate it */
    int status;
    pid_t result = waitpid(pid, &status, 0);
    if (result < 0) {
        perror(name.c_str());
    } else if (WIFEXITED(status)) {
        printf("session %d [%d]: exited with code %d\n", int(session_id), int(pid), int(WEXITSTATUS(status)));
    } else if (WIFSIGNALED(status)) {
        printf("session %d [%d]: terminated by signal %d\n", int(session_id), int(pid), int(WTERMSIG(status)));
    } else {
        printf("session %d [%d]: unknown termination code 0x%04X\n", int(session_id), int(pid), status);
    }
    pid = 0;
    readfd = writefd = -1;

    /* Notify file server */
    maybeNotifyFileServer();
}

/** Restart session.
    \param server Name of server binary
    \return true on success, false on failure */
bool
RouterSession::restart(const char* server)
{
    stop();
    answer.clear();
    return start(server);
}

/** Execute request. This transfers data to and from the child, and writes
    the protocol response to the socket fd.
    \param req Partial request object
    \param fd Socket fd */
void
RouterSession::executeRequest(Request& req, int fd)
{
    /* Read one line */
    char* command;
    while ((command = req.readLine()) == 0) {
        if (req.readMore(fd) != Request::sSuccess) {
            writeStr(fd, "400 Bad Request\n");
            return;
        }
    }

    /* Log it */
    printf("session %d [%d]: '%s'\n", int(session_id), int(pid), command);
    used = true;
    modified = !isSAVE(command);

    /* Send line to child */
    writeStr(writefd, command);
    write(writefd, "\n", 1);
    if (isPOST(command)) {
        while (1) {
            char* line = req.readLine();
            if (line == 0) {
                /* No complete line; try reading some more data */
                Request::State st = req.readMore(fd);
                if (st != Request::sSuccess
                    && !(st == Request::sTooLong && req.shift()))
                {
                    /* Read not successful. However, we cannot do anything
                       sensible about it. Let the server fail it. */
                    break;
                }
            } else if (std::strcmp(line, ".") == 0) {
                /* End marker */
                break;
            } else {
                writeStr(writefd, line);
                write(writefd, "\n", 1);
            }
        }
        write(writefd, ".\n", 2);
    }

    /* Read reply from child */
    answer.clear();
    char* reply;
    while ((reply = answer.readLine()) == 0) {
        Request::State st = answer.readMore(readfd);
        if (st == Request::sEOF) {
            writeStr(fd, "452 Session timeout (unexpected)\n");
            stop();
            return;
        } else if (st != Request::sSuccess) {
            writeStr(fd, "500 Invalid answer from server\n");
            return;
        }
    }
    writeStr(fd, reply);
    write(fd, "\r\n", 2);

    /* Copy request body */
    /* FIXME: this is not too efficient */
    if (reply[0] == '2') {
        while (1) {
            if ((reply = answer.readLine()) == 0) {
                /* No complete line; try reading some more data */
                Request::State st = answer.readMore(readfd);
                if (st != Request::sSuccess
                    && !(st == Request::sTooLong && answer.shift()))
                {
                    /* Read not successful, and it's not a buffer size problem we can fix */
                    if (st == Request::sError) {
                        printf("session %d [%d]: request terminated with error '%s'\n", int(session_id), int(pid), strerror(errno));
                    } else if (st == Request::sTooLong) {
                        printf("session %d [%d]: request terminated with overlong line\n", int(session_id), int(pid));
                    } else if (st == Request::sEOF) {
                        printf("session %d [%d]: request terminated with unexpected EOF\n", int(session_id), int(pid));
                    } else {
                        printf("session %d [%d]: request terminated unexpectedly\n", int(session_id), int(pid));
                    }
                    break;
                }
            } else if (std::strcmp(reply, ".") == 0) {
                /* Terminator. Finish this request. */
                break;
            } else {
                /* Line for client */
                errno = 0;
                writeStr(fd, reply);
                int b = write(fd, "\r\n", 2);
                if (b != 2) {
                    printf("session %d [%d]: write error, '%s'\n", int(session_id), int(pid), strerror(errno));
                    break;
                }
            }
        }
    }

    last_access = time(0);
}

/** Save this session. */
bool
RouterSession::save(bool notify)
{
    /* Bureaucracy */
    last_access = time(0);
    used = true;
    if (!modified) {
        return false;
    }

    /* When we're here, the session was modified and can be saved */
    printf("session %d [%d]: 'SAVE' (from router)\n", int(session_id), int(pid));
    writeStr(writefd, "SAVE\n");

    /* Read reply */
    answer.clear();
    char* reply;
    while ((reply = answer.readLine()) == 0) {
        Request::State st = answer.readMore(readfd);
        if (st == Request::sEOF) {
            /* Error: unexpected termination */
            stop();
            return true;
        } else if (st != Request::sSuccess) {
            /* Error: invalid answer from server */
            return true;
        }
    }

    /* Normally, we would have to skip a possible reply body here, but
       SAVE answers with a '100' reply without body. So all we need to
       do is to register that this session is now saved. */
    modified = false;

    /* Notify file server */
    if (notify) {
        maybeNotifyFileServer();
    }
    return true;
}
